import json
import logging
from pathlib import Path
from typing import Optional

import typer
from typing_extensions import Annotated
from rich import print

from yamlex.mapper import (
    validate_json_schemas_dir,
    map_schema_to_sources,
    read_vscode_settings_json_file,
    extract_definitions_into_standalone_schemas,
)
from yamlex.splitter import split_yaml
from yamlex.joiner import (
    assemble_recursively,
    remove_yaml_comments,
)
from yamlex.util import (
    adjust_root_logger,
    get_default_extension_dir_path,
    get_default_extension_source_dir_path,
    is_manually_created,
    write_file_with_generated_comment,
    read_version,
    write_version,
)


logging.basicConfig(level=logging.INFO, format="%(message)s")

logger = logging.getLogger(__name__)
app = typer.Typer(
    name="yamlex",
    rich_markup_mode="rich",
    no_args_is_help=True,
    add_completion=False,
)

force_option = Annotated[
    bool,
    typer.Option(
        "--force",
        "-f",
        help="Overwrite the files even if they were created manually.",
    ),
]
verbose_option = Annotated[
    bool,
    typer.Option(
        "--verbose",
        help="Enable verbose output.",
    ),
]
quiet_option = Annotated[
    bool,
    typer.Option(
        "--quiet",
        help="Disable any informational output. Only errors.",
    ),
]
debug_option = Annotated[
    bool,
    typer.Option(
        "--debug",
        hidden=True,
    )
]


@app.command(help="Map JSON schema to YAML files in VS Code settings")
def map(
    settings: Annotated[
        Path,
        typer.Argument(
            help="Path to the VS Code settings.json file.",
        ),
    ] = Path(".vscode/settings.json"),
    schema: Annotated[
        Path,
        typer.Option(
            "--json",
            "-j",
            help="Path to directory with valid extensions JSON schema files.",
            exists=True,
            readable=True,
            dir_okay=True,
            file_okay=False,
        ),
    ] = Path("schema"),
    source: Annotated[
        Optional[Path],
        typer.Option(
            "--source",
            "-s",
            help=(
                "Path to directory where YAML source files will be stored. "
                "[dim]\\[default: source or src/source][/dim]"
            ),
            show_default=False,
            dir_okay=True,
            file_okay=False,
        )
    ] = None,
    root: Annotated[
        Path,
        typer.Option(
            "--root",
            "-r",
            help="Root directory relative to which the paths in settings file will be mapped.",
            dir_okay=True,
            file_okay=False,
        )
    ] = Path("."),
    extension_yaml: Annotated[
        Optional[Path],
        typer.Option(
            "--extension-yaml",
            "-e",
            help=(
                "Path to output extension.yaml file. "
                "[dim]\\[default: extension/extension.yaml or src/extension/extension.yaml][/dim]"
            ),
            show_default=False,
            dir_okay=False,
            file_okay=True,
        )
    ] = None,
    verbose: verbose_option = False,
    quiet: quiet_option = False,
):
    """Map YAML files to JSON schema."""
    adjust_root_logger(verbose, quiet)
    logger.debug(f"JSON schema files directory: {schema}")
    logger.debug(f"VS Code settings.json file: {settings}")
    logger.debug(f"Root dir: {root}")

    source = source or get_default_extension_source_dir_path()
    logger.debug(f"Extension YAML source files directory: {source}")

    extension_yaml = extension_yaml or get_default_extension_dir_path() / "extension.yaml"
    logger.debug(f"extension.yaml file: {extension_yaml}")

    # Make sure schemas directory contains extension.schema.json
    validate_json_schemas_dir(schema)

    # Make schema more granular by extracting embedded definitions into
    # separate schema files.
    extract_definitions_into_standalone_schemas(schema)

    # Update yaml.schemas mapping in settings.json
    vscode_settings = read_vscode_settings_json_file(settings)

    # Delete existing extension.yaml mapping
    mapping: dict = vscode_settings.get("yaml.schemas", {})
    for k, v in mapping.items():
        if "extension.yaml" in str(v):
            del mapping[k]
            break
    
    # Update the mapping
    mapping_update = map_schema_to_sources(schema, source, root, extension_yaml)
    mapping.update(mapping_update)

    # Write the updated settings.json file
    vscode_settings["yaml.schemas"] = mapping
    with open(settings, "w") as f:
        vscode_settings_file_text = json.dumps(vscode_settings, indent=2)
        f.write(vscode_settings_file_text)
        logger.info(f"Updated {settings} with new YAML schema mappings.")


@app.command(help="Split central YAML file into parts.")
def split(
    source: Annotated[
        Optional[Path],
        typer.Option(
            "--source",
            "-s",
            help=(
                "Path to source extension.yaml file. "
                "[dim]\\[default: extension/extension.yaml or src/extension/extension.yaml][/dim]"
            ),
            show_default=False,
            dir_okay=False,
            file_okay=True,
            exists=True,
            readable=True,
        )
    ] = None,
    target: Annotated[
        Optional[Path],
        typer.Option(
            "--target",
            "-t",
            help=(
                "Path to directory where split YAML source files will be stored. "
                "[dim]\\[default: source or src/source][/dim]"
            ),
            show_default=False,
            dir_okay=True,
            file_okay=False,
            exists=True,
            writable=True,
        )
    ] = None,
    force: force_option = False,
    verbose: verbose_option = False,
    quiet: quiet_option = False,
):
    adjust_root_logger(verbose, quiet)

    source = source or get_default_extension_dir_path() / "extension.yaml"
    logger.debug(f"Source file: {source}")

    target = target or get_default_extension_source_dir_path()
    logger.debug(f"Target directory: {target}")

    split_parts = split_yaml(source, target)

    for path, part in split_parts.items():
        if is_manually_created(path) and not force:
            logger.info(f"(Skipping) Part: {path}")
        else:
            write_file_with_generated_comment(path, part)
            logger.info(f"Part written: {path}")


@app.command(help="Join YAML files into a single file.")
def join(
    source: Annotated[
        Optional[Path],
        typer.Option(
            "--source",
            "-s",
            help=(
                "Path to directory where split YAML source files are stored. "
                "[dim]\\[default: csource or src/source][/dim]"
            ),
            show_default=False,
            dir_okay=True,
            file_okay=False,
            exists=True,
            readable=True,
        )
    ] = None,
    target: Annotated[
        Optional[Path],
        typer.Option(
            "--target",
            "-t",
            help=(
                "Path for target extension.yaml file that will be assembled from parts. "
                "[dim]\\[default: extension/extension.yaml or src/extension/extension.yaml][/dim]"
            ),
            show_default=False,
            dir_okay=False,
            file_okay=True,
        )
    ] = None,
    dev: Annotated[
        bool,
        typer.Option(
            "--dev",
            "-d",
            help="Prefix extension name with 'custom:' and use explicit version.",
        ),
    ] = False,
    bump: Annotated[
        bool,
        typer.Option(
            "--bump",
            "-b",
            help="Bump version in the version.properties file.",
        )
    ] = False,
    version: Annotated[
        Optional[str],
        typer.Option(
            "--version",
            "-v",
            help="Explicitly set the version to use during assembly or bump process."
        ),
    ] = None,
    force: force_option = False,
    verbose: verbose_option = False,
    quiet: quiet_option = False,
    debug: debug_option = False,
):
    adjust_root_logger(verbose, quiet)
    
    # Bump version, if requested
    if bump:
        current_version = read_version("version.properties")
        logger.info(f"Current version: {current_version}")
        # If new explicit version is not specified, then
        # parse current version and bump it.
        if not version:
            version_parts = current_version.split(".")
            part_int: int | None = None
            part_idx: int | None = None
            for i, part in enumerate(reversed(version_parts)):
                try:
                    part_int = int(part)
                    part_idx = len(version_parts) - 1 - i
                    break
                except:
                    continue
            if not part_int:
                logger.error("Could not find integer part in version to bump")
                exit(1)

            part_int += 1
            version = ""
            for i, part in enumerate(version_parts):
                append = part
                if i == part_idx:
                    append = part_int
                version += f"{append}."
            version = version.rstrip(".")

        write_version("version.properties", version)
        logger.info(f"Bumped version to {version}")

    source = source or get_default_extension_source_dir_path()
    logger.debug(f"Source files directory: {source}")

    extension = assemble_recursively(source, debug=debug)

    if dev:
        if not version:
            version = read_version("version.properties")

        name = extension.get("name")
        if isinstance(name, str) and not name.startswith("custom:"):
            name = f"custom:{name}"
            extension["name"] = name
            logger.info(f"Set extension name to: {name}")

        extension["version"] = version
        logger.info(f"Set extension version to: {version}")

    target = target or get_default_extension_dir_path() / "extension.yaml"
    logger.debug(f"Target file: {target}")

    # Check if the target file exists and was created manually
    if is_manually_created(target) and not force:
        logger.error(f"The {target} file was created manually. Use --force to overwrite it.")
        exit(2)

    # Remove comments
    remove_yaml_comments(extension)

    # Write to output file
    write_file_with_generated_comment(target, extension)

    if debug:
        print(extension)


def run():
    app.command(name="j", hidden=True)(join)
    app.command(name="s", hidden=True)(split)
    app()



import logging
from pathlib import Path
from typing import Union

import ruamel.yaml


logger = logging.getLogger(__name__)
parser = ruamel.yaml.YAML()


# How joiner works
# - The goal is to recreate the YAML structure of the extension.yaml from files
# - All files with *.yaml extension are considered
# - We start to rebuild from the same level where extension.yaml is. Loop:


def assemble_recursively(dir_path: Path, debug: bool = False) -> Union[dict, list]:
    # List all files
    all_paths = [p for p in dir_path.glob("*") if not p.is_symlink()]
    all_dirs: list[Path] = []
    all_yamls: dict[str, Path] = {}
    other_files: dict[str, Path] = {}
    for p in all_paths:
        if p.is_dir():
            all_dirs.append(p)
        elif p.is_file():
            if p.suffix in (".yaml", ".yml"):
                all_yamls[p.stem] = p
            else:
                other_files[p.stem] = p

    # Figure out if the current folder is an array
    is_array = any([p for p in all_paths if p.name.startswith("-")])
    logger.debug(f"Assembling {'(array)' if is_array else ''} level: {dir_path}")

    # All parsed data from index file, other yaml files, and dirs
    # will be collected into a single data object.
    data: Union[dict, list] = [] if is_array else {}
    
    # Load data from the index file, if it exists
    if not is_array:
        if "index" in all_yamls:
            index_yaml_file_path = all_yamls.pop("index")
            with open(index_yaml_file_path, "r") as index_yaml_file:
                # index_yaml_file_data = yaml.safe_load(index_yaml_file)
                index_yaml_file_data = parser.load(index_yaml_file)
                data.update(index_yaml_file_data)

    # Load data from the rest of the YAML files.
    # File's name is considered to be the name of the nested field,
    # if it's not an array. Otherwise, file name is ignored completely
    for yaml_file_name, yaml_file_path in all_yamls.items():
        with open(yaml_file_path, "r") as yaml_file:
            # yaml_file_data = yaml.safe_load(yaml_file)
            yaml_file_data = parser.load(yaml_file)

            if is_array:
                data.append(yaml_file_data)
            else:
                data[yaml_file_name] = yaml_file_data

    # Load data from the rest of the files.
    # Example: query.sql file containing a SQL query
    for other_file_name, other_file_path in other_files.items():
        with open(other_file_path, "r") as other_file:
            other_file_content = other_file.read()

            if is_array:
                data.append(other_file_content)
            else:
                data[other_file_name] = other_file_content 

    # Recursively traverse directories.
    # Directorie's name is considered to be the name of the nested field,
    # if it's not an array. Otherwise, directory name is ignored.
    for dir in all_dirs:
        dir_data = assemble_recursively(dir, debug=debug)

        # If directorie's name starts with a '+', then consider it a grouper
        # It doesn't represent a field. Consider data within it to be on the
        # same level as the current folder.
        # Does not apply if the current folder is considered to be an array.
        is_grouper = dir.name.startswith("+")

        if is_array:
            data.append(dir_data)
        else:
            if is_grouper:
                data.update(dir_data)
            else:
                data[dir.name] = dir_data

    return data


def remove_yaml_comments(data: Union[dict, list]):
    if hasattr(data, 'ca'):
        # Remove comments associated with this node
        data.ca.comment = None
        
        if hasattr(data.ca.items, 'items'):
            for key in data.ca.items:
                # Remove key comments
                data.ca.items[key] = [None, None, None, None]

    if isinstance(data, dict):
        for value in data.values():
            remove_yaml_comments(value)
            
    elif isinstance(data, list):
        for item in data:
            remove_yaml_comments(item)



import logging
import re
from datetime import datetime
from io import StringIO
from pathlib import Path
from typing import Optional, Union

import ruamel.yaml


logger = logging.getLogger(__name__)
parser = ruamel.yaml.YAML()


def adjust_root_logger(verbose: bool = False, quiet: bool = False) -> None:
    if quiet:
        logging.root.setLevel(logging.ERROR)
    elif verbose:
        logging.root.setLevel(logging.DEBUG)


def get_default_extension_source_dir_path() -> Path:
    src_dir_path = Path("src")
    if src_dir_path.exists() and src_dir_path.is_dir():
        return src_dir_path / "source"
    else:
        return Path("source")


def get_default_extension_dir_path() -> Path:
    candidates = [
        Path("extension"),
        Path("src/extension"),
    ]

    # Find the first suitable candidate
    for candidate in candidates:
        # Check the candidate directory exists
        if candidate.exists() and candidate.is_dir():
            return candidate
        continue

    # Nothing suitable was found
    logger.error(
        (
            "No default extension directory found. "
            "'extension/' or 'src/extension/' folder must exist "
            "in the current working directory."
        )
    )
    exit(1)


def is_manually_created(path: Path) -> bool:
    if path.exists():
        with open(path, "r") as f:
            content = f.read()
            if "generated by yamlex" not in content:
                True
    return False       


def get_generated_content_comment() -> str:
    comment = (
        "# This file was generated by yamlex\n"
        f"# at: {datetime.now().isoformat()}\n"
        "\n"
    )
    return comment


def write_file_with_generated_comment(
    file_path: Path,
    data: Union[dict, list],
) -> None:
    # Convert dict to YAML. Dump to string first to add a comment
    stream = StringIO()
    parser.indent(mapping=2, sequence=4, offset=2)
    parser.dump(data, stream)
    text = stream.getvalue()

    # Make sure the directory we write to exists
    dir = file_path.parent.resolve()
    dir.mkdir(parents=True, exist_ok=True)

    # Write to output file
    with open(file_path, "w") as f:
        # Write a comment to indicate that the file was automatically generated
        comment = get_generated_content_comment()
        f.write(comment)
        f.write(text)


def read_version(path: Path, default: Optional[str] = None) -> str:
    try:
        with open(path, "r") as f:
            content = f.read()
            match_ = re.search(r"version=(.*)", content)
            if not match_ and not default:
                raise KeyError("Could not read current version")
            return match_.group(1)
    except Exception as e:
        logger.error(f"Failed to open {path}: {e}")
        exit(1)


def write_version(path: Path, version: str) -> None:
    try:
        with open(path, "r") as rf:
            lines = rf.readlines()
            content = []
            for l in lines:
                if l.startswith("version"):
                    content.append(f"version={version}\n")
                else:
                    content.append(l)
        with open(path, "w") as wf:
            wf.writelines(content) 
    except Exception as e:
        logger.error(f"Failed to write to {path}: {e}")
        exit(1)
